@@ -41,6 +41,13 @@ class QoS(IntEnum):
     Note that this client does not currently support QoS 2 as of (August 2022)
     """
 
+    def to_mqtt3(self):
+        from awscrt.mqtt import QoS as Mqtt3QoS
+        """Convert a Mqtt5 QoS to Mqtt3
+
+        """
+        return Mqtt3QoS(self.value)
+
 
 def _try_qos(value):
     try:
@@ -899,6 +906,72 @@ def _init_user_properties(user_properties_tuples):
     return [UserProperty(name=name, value=value) for (name, value) in user_properties_tuples]
 
 
+class OutboundTopicAliasBehaviorType(IntEnum):
+    """An enumeration that controls how the client applies topic aliasing to outbound publish packets.
+
+    Topic alias behavior is described in `MQTT5 Topic Aliasing <https://docs.oasis-open.org/mqtt/mqtt/v5.0/os/mqtt-v5.0-os.html#_Toc3901113>`_
+    """
+
+    DEFAULT = 0,
+    """Maps to Disabled.  This keeps the client from being broken (by default) if the broker
+     topic aliasing implementation has a problem.
+    """
+
+    MANUAL = 1,
+    """
+     Outbound aliasing is the user's responsibility.  Client will cache and use
+     previously-established aliases if they fall within the negotiated limits of the connection.
+
+     The user must still always submit a full topic in their publishes because disconnections disrupt
+     topic alias mappings unpredictably.  The client will properly use a requested alias when the most-recently-seen
+     binding for a topic alias value matches the alias and topic in the publish packet.
+    """
+
+    LRU = 2,
+    """ (Recommended) The client will ignore any user-specified topic aliasing and instead use an LRU cache to drive
+    alias usage.
+    """
+
+    DISABLED = 3,
+    """Completely disable outbound topic aliasing."""
+
+
+class InboundTopicAliasBehaviorType(IntEnum):
+    """An enumeration that controls whether or not the client allows the broker to send publishes that use topic
+        aliasing.
+
+    Topic alias behavior is described in `MQTT5 Topic Aliasing <https://docs.oasis-open.org/mqtt/mqtt/v5.0/os/mqtt-v5.0-os.html#_Toc3901113>`_
+    """
+
+    DEFAULT = 0,
+    """Maps to Disabled.  This keeps the client from being broken (by default) if the broker
+     topic aliasing implementation has a problem.
+     """
+
+    ENABLED = 1,
+    """Allow the server to send PUBLISH packets to the client that use topic aliasing"""
+
+    DISABLED = 2,
+    """Forbid the server from sending PUBLISH packets to the client that use topic aliasing"""
+
+
+@dataclass
+class TopicAliasingOptions:
+    """
+    Configuration for all client topic aliasing behavior.
+
+    Args:
+        outbound_behavior (OutboundTopicAliasBehaviorType):  Controls what kind of outbound topic aliasing behavior the client should attempt to use.  If topic aliasing is not supported by the server, this setting has no effect and any attempts to directly manipulate the topic alias id in outbound publishes will be ignored.  If left undefined, then outbound topic aliasing is disabled.
+        outbound_cache_max_size (int):  If outbound topic aliasing is set to LRU, this controls the maximum size of the cache.  If outbound topic aliasing is set to LRU and this is zero or undefined, a sensible default is used (25).  If outbound topic aliasing is not set to LRU, then this setting has no effect.
+        inbound_behavior (InboundTopicAliasBehaviorType):  Controls whether or not the client allows the broker to use topic aliasing when sending publishes.  Even if inbound topic aliasing is enabled, it is up to the server to choose whether or not to use it.  If left undefined, then inbound topic aliasing is disabled.
+        inbound_cache_max_size (int):  If inbound topic aliasing is enabled, this will control the size of the inbound alias cache.  If inbound aliases are enabled and this is zero or undefined, then a sensible default will be used (25).  If inbound aliases are disabled, this setting has no effect.  Behaviorally, this value overrides anything present in the topic_alias_maximum field of the CONNECT packet options.
+    """
+    outbound_behavior: OutboundTopicAliasBehaviorType = None
+    outbound_cache_max_size: int = None
+    inbound_behavior: InboundTopicAliasBehaviorType = None
+    inbound_cache_max_size: int = None
+
+
 @dataclass
 class NegotiatedSettings:
     """
@@ -918,9 +991,9 @@ class NegotiatedSettings:
         session_expiry_interval_sec (int): The amount of time in seconds the server will retain the MQTT session after a disconnect.
         receive_maximum_from_server (int): The number of in-flight QoS 1 and QoS 2 publications the server is willing to process concurrently.
         maximum_packet_size_to_server (int): The maximum packet size the server is willing to accept.
-        topic_alias_maximum_to_server (int): Not Currently Supported
-        topic_alias_maximum_to_client (int): Not Currently Supported
-        server_keep_alive_sec (int): The maximum amount of time in seconds between client packets. The client should use PINGREQs to ensure this limit is not breached.  The server will disconnect the client for inactivity if no MQTT packet is received in a time interval equal to 1.5 x this value.
+        topic_alias_maximum_to_server (int): the maximum allowed topic alias value on publishes sent from client to server
+        topic_alias_maximum_to_client (int): the maximum allowed topic alias value on publishes sent from server to client
+        server_keep_alive_sec (int): The maximum amount of time in seconds between client packets. The client will use PINGREQs to ensure this limit is not breached.  The server will disconnect the client for inactivity if no MQTT packet is received in a time interval equal to 1.5 x this value.
         retain_available (bool): Whether the server supports retained messages.
         wildcard_subscriptions_available (bool): Whether the server supports wildcard subscriptions.
         subscription_identifiers_available (bool): Whether the server supports subscription identifiers
@@ -953,6 +1026,7 @@ class ConnackPacket:
         retain_available (bool): Indicates whether the server supports retained messages.  If None, retained messages are supported.
         maximum_packet_size (int): Specifies the maximum packet size, in bytes, that the server is willing to accept.  If None, there is no limit beyond what is imposed by the MQTT spec itself.
         assigned_client_identifier (str): Specifies a client identifier assigned to this connection by the server.  Only valid when the client id of the preceding CONNECT packet was left empty.
+        topic_alias_maximum (int): The maximum allowed value for topic aliases in outbound publish packets.  If 0 or None, then outbound topic aliasing is not allowed.
         reason_string (str): Additional diagnostic information about the result of the connection attempt.
         user_properties (Sequence[UserProperty]): List of MQTT5 user properties included with the packet.
         wildcard_subscriptions_available (bool): Indicates whether the server supports wildcard subscriptions.  If None, wildcard subscriptions are supported.
@@ -970,6 +1044,7 @@ class ConnackPacket:
     retain_available: bool = None
     maximum_packet_size: int = None
     assigned_client_identifier: str = None
+    topic_alias_maximum: int = None
     reason_string: str = None
     user_properties: 'Sequence[UserProperty]' = None
     wildcard_subscriptions_available: bool = None
@@ -1083,7 +1158,7 @@ class PublishPacket:
         topic (str): The topic associated with this PUBLISH packet.
         payload_format_indicator (PayloadFormatIndicator): Property specifying the format of the payload data. The mqtt5 client does not enforce or use this value in a meaningful way.
         message_expiry_interval_sec (int): Sent publishes - indicates the maximum amount of time allowed to elapse for message delivery before the server should instead delete the message (relative to a recipient). Received publishes - indicates the remaining amount of time (from the server's perspective) before the message would have been deleted relative to the subscribing client. If left None, indicates no expiration timeout.
-        topic_alias (int): An integer value that is used to identify the Topic instead of using the Topic Name.
+        topic_alias (int): An integer value that is used to identify the Topic instead of using the Topic Name.  On outbound publishes, this will only be used if the outbound topic aliasing behavior has been set to Manual.
         response_topic (str): Opaque topic string intended to assist with request/response implementations.  Not internally meaningful to MQTT5 or this client.
         correlation_data (Any): Opaque binary data used to correlate between publish messages, as a potential method for request-response implementation.  Not internally meaningful to MQTT5.
         subscription_identifiers (Sequence[int]): The subscription identifiers of all the subscriptions this message matched.
@@ -1308,6 +1383,7 @@ class ClientOptions:
         ping_timeout_ms (int): The time interval to wait after sending a PINGREQ for a PINGRESP to arrive. If one does not arrive, the client will close the current connection.
         connack_timeout_ms (int): The time interval to wait after sending a CONNECT request for a CONNACK to arrive.  If one does not arrive, the connection will be shut down.
         ack_timeout_sec (int): The time interval to wait for an ack after sending a QoS 1+ PUBLISH, SUBSCRIBE, or UNSUBSCRIBE before failing the operation.
+        topic_aliasing_options (TopicAliasingOptions): All configurable options with respect to client topic aliasing behavior.
         on_publish_callback_fn (Callable[[PublishReceivedData],]): Callback for all publish packets received by client.
         on_lifecycle_event_stopped_fn (Callable[[LifecycleStoppedData],]): Callback for Lifecycle Event Stopped.
         on_lifecycle_event_attempting_connect_fn (Callable[[LifecycleAttemptingConnectData],]): Callback for Lifecycle Event Attempting Connect.
@@ -1333,6 +1409,7 @@ class ClientOptions:
     ping_timeout_ms: int = None
     connack_timeout_ms: int = None
     ack_timeout_sec: int = None
+    topic_aliasing_options: TopicAliasingOptions = None
     on_publish_callback_fn: Callable[[PublishReceivedData], None] = None
     on_lifecycle_event_stopped_fn: Callable[[LifecycleStoppedData], None] = None
     on_lifecycle_event_attempting_connect_fn: Callable[[LifecycleAttemptingConnectData], None] = None
@@ -1456,6 +1533,8 @@ class _ClientCore:
             connack_maximum_packet_size_exists,
             connack_maximum_packet_size,
             connack_assigned_client_identifier,
+            connack_topic_alias_maximum_exists,
+            connack_topic_alias_maximum,
             connack_reason_string,
             connack_user_properties_tuples,
             connack_wildcard_subscriptions_available_exist,
@@ -1497,6 +1576,8 @@ class _ClientCore:
         if connack_maximum_packet_size_exists:
             connack_packet.maximum_packet_size = connack_maximum_packet_size
         connack_packet.assigned_client_identifier = connack_assigned_client_identifier
+        if connack_topic_alias_maximum_exists:
+            connack_packet.topic_alias_maximum = connack_topic_alias_maximum
         connack_packet.reason_string = connack_reason_string
         connack_packet.user_properties = _init_user_properties(connack_user_properties_tuples)
         if connack_wildcard_subscriptions_available_exist:
@@ -1630,6 +1711,48 @@ class _ClientCore:
                     exception=exceptions.from_code(error_code)))
 
 
+@dataclass
+class _Mqtt5to3AdapterOptions:
+    """This internal class stores the options that required for creating a new Mqtt3 connection from the mqtt5 client
+    Args:
+        host_name (str): Host name of the MQTT server to connect to.
+        port (int): Network port of the MQTT server to connect to.
+        client_id (str): A unique string identifying the client to the server.  Used to restore session state between connections. If left empty, the broker will auto-assign a unique client id.  When reconnecting, the mqtt5 client will always use the auto-assigned client id.
+        socket_options (SocketOptions): The socket properties of the underlying MQTT connections made by the client or None if defaults are used.
+        min_reconnect_delay_ms (int): The minimum amount of time to wait to reconnect after a disconnect. Exponential backoff is performed with jitter after each connection failure.
+        max_reconnect_delay_ms (int): The maximum amount of time to wait to reconnect after a disconnect.  Exponential backoff is performed with jitter after each connection failure.
+        ping_timeout_ms (int): The time interval to wait after sending a PINGREQ for a PINGRESP to arrive. If one does not arrive, the client will close the current connection.
+        keep_alive_secs (int): The keep alive value, in seconds, A PING will automatically be sent at this interval.
+        ack_timeout_secs (int): The time interval to wait for an ack after sending a QoS 1+ PUBLISH, SUBSCRIBE, or UNSUBSCRIBE before failing the operation.
+        clean_session (bool): Whether or not to start a clean session with each reconnect.
+
+        The default values are referred from awscrt.mqtt.Connection
+    """
+
+    def __init__(
+            self,
+            host_name: str,
+            port: int,
+            client_id: str,
+            socket_options: SocketOptions,
+            min_reconnect_delay_ms: int,
+            max_reconnect_delay_ms: int,
+            ping_timeout_ms: int,
+            keep_alive_secs: int,
+            ack_timeout_secs: int,
+            clean_session: int):
+        self.host_name = host_name
+        self.port = port
+        self.client_id = "" if client_id is None else client_id
+        self.socket_options = socket_options
+        self.min_reconnect_delay_ms = 5 if min_reconnect_delay_ms is None else min_reconnect_delay_ms
+        self.max_reconnect_delay_ms: int = 60 if max_reconnect_delay_ms is None else max_reconnect_delay_ms
+        self.ping_timeout_ms: int = 3000 if ping_timeout_ms is None else ping_timeout_ms
+        self.keep_alive_secs: int = 1200 if keep_alive_secs is None else keep_alive_secs
+        self.ack_timeout_secs: int = 0 if ack_timeout_secs is None else ack_timeout_secs
+        self.clean_session: bool = True if clean_session is None else clean_session
+
+
 class Client(NativeResource):
     """This class wraps the aws-c-mqtt MQTT5 client to provide the basic MQTT5 pub/sub functionalities via the AWS Common Runtime
 
@@ -1667,7 +1790,7 @@ class Client(NativeResource):
             will = connect_options.will
 
         websocket_is_none = client_options.websocket_handshake_transform is None
-
+        self.tls_ctx = client_options.tls_ctx
         self._binding = _awscrt.mqtt5_client_new(self,
                                                  client_options.host_name,
                                                  client_options.port,
@@ -1707,9 +1830,24 @@ class Client(NativeResource):
                                                  client_options.min_connected_time_to_reset_reconnect_delay_ms,
                                                  client_options.ping_timeout_ms,
                                                  client_options.ack_timeout_sec,
+                                                 client_options.topic_aliasing_options,
                                                  websocket_is_none,
                                                  core)
 
+        # Store the options for adapter
+        self.adapter_options = _Mqtt5to3AdapterOptions(
+            host_name=client_options.host_name,
+            port=client_options.port,
+            client_id=connect_options.client_id,
+            socket_options=socket_options,
+            min_reconnect_delay_ms=client_options.min_reconnect_delay_ms,
+            max_reconnect_delay_ms=client_options.max_reconnect_delay_ms,
+            ping_timeout_ms=client_options.ping_timeout_ms,
+            keep_alive_secs=connect_options.keep_alive_interval_sec,
+            ack_timeout_secs=client_options.ack_timeout_sec,
+            clean_session=(
+                client_options.session_behavior < ClientSessionBehaviorType.REJOIN_ALWAYS if client_options.session_behavior else True))
+
     def start(self):
         """Notifies the MQTT5 client that you want it maintain connectivity to the configured endpoint.
         The client will attempt to stay connected using the properties of the reconnect-related parameters in the mqtt5 client configuration.
@@ -1848,23 +1986,97 @@ class Client(NativeResource):
         """Queries the client's internal statistics for incomplete operations.
 
         Returns:
-            A future with a (:class:`OperationStatisticsData`)
+            The (:class:`OperationStatisticsData`) containing the statistics
         """
 
-        future = Future()
+        result = _awscrt.mqtt5_client_get_stats(self._binding)
+        return OperationStatisticsData(result[0], result[1], result[2], result[3])
 
-        def get_stats_result(
-                incomplete_operation_count,
-                incomplete_operation_size,
-                unacked_operation_count,
-                unacked_operation_size):
-            operation_statistics_data = OperationStatisticsData(
-                incomplete_operation_count,
-                incomplete_operation_size,
-                unacked_operation_count,
-                unacked_operation_size)
-            future.set_result(operation_statistics_data)
+    def new_connection(self, on_connection_interrupted=None, on_connection_resumed=None,
+                       on_connection_success=None, on_connection_failure=None, on_connection_closed=None):
+        from awscrt.mqtt import Connection
+        """ Returns a new Mqtt3 Connection Object wraps the Mqtt5 client.
 
-        _awscrt.mqtt5_client_get_stats(self._binding, get_stats_result)
+            Args:
+                on_connection_interrupted: Optional callback invoked whenever the MQTT connection is lost.
+                    The MQTT client will automatically attempt to reconnect.
+                    The function should take the following arguments return nothing:
 
-        return future
+                        *   `connection` (:class:`Connection`): This MQTT Connection.
+
+                        *   `error` (:class:`awscrt.exceptions.AwsCrtError`): Exception which caused connection loss.
+
+                        *   `**kwargs` (dict): Forward-compatibility kwargs.
+
+                on_connection_resumed: Optional callback invoked whenever the MQTT connection
+                    is automatically resumed. Function should take the following arguments and return nothing:
+
+                        *   `connection` (:class:`Connection`): This MQTT Connection
+
+                        *   `return_code` (:class:`ConnectReturnCode`): Connect return
+                            code received from the server.
+
+                        *   `session_present` (bool): True if resuming existing session. False if new session.
+                            Note that the server has forgotten all previous subscriptions if this is False.
+                            Subscriptions can be re-established via resubscribe_existing_topics().
+
+                        *   `**kwargs` (dict): Forward-compatibility kwargs.
+
+                on_connection_success: Optional callback invoked whenever the connection successfully connects.
+                    This callback is invoked for every successful connect and every successful reconnect.
+
+                    Function should take the following arguments and return nothing:
+
+                        * `connection` (:class:`Connection`): This MQTT Connection
+
+                        * `callback_data` (:class:`OnConnectionSuccessData`): The data returned from the connection success.
+
+                on_connection_failure: Optional callback invoked whenever the connection fails to connect.
+                    This callback is invoked for every failed connect and every failed reconnect.
+
+                    Function should take the following arguments and return nothing:
+
+                        * `connection` (:class:`Connection`): This MQTT Connection
+
+                        * `callback_data` (:class:`OnConnectionFailureData`): The data returned from the connection failure.
+
+                on_connection_closed: Optional callback invoked whenever the connection has been disconnected and shutdown successfully.
+                    Function should take the following arguments and return nothing:
+
+                    * `connection` (:class:`Connection`): This MQTT Connection
+
+                    * `callback_data` (:class:`OnConnectionClosedData`): The data returned from the connection close.
+
+
+            Returns:
+                The (:class:`Connection`) wrapper for the mqtt5 client
+        """
+        return Connection(
+            self,
+            self.adapter_options.host_name,
+            self.adapter_options.port,
+            self.adapter_options.client_id,
+            clean_session=self.adapter_options.clean_session,
+            on_connection_interrupted=on_connection_interrupted,
+            on_connection_resumed=on_connection_resumed,
+            on_connection_success=on_connection_success,
+            on_connection_failure=on_connection_failure,
+            on_connection_closed=on_connection_closed,
+            reconnect_min_timeout_secs=self.adapter_options.min_reconnect_delay_ms,
+            reconnect_max_timeout_secs=self.adapter_options.max_reconnect_delay_ms,
+            keep_alive_secs=self.adapter_options.keep_alive_secs,
+            ping_timeout_ms=self.adapter_options.ping_timeout_ms,
+            protocol_operation_timeout_ms=self.adapter_options.ack_timeout_secs * 1000,
+            socket_options=self.adapter_options.socket_options,
+
+            # For the arugments below, set it to `None` will directly use the options from mqtt5 client underlying.
+            will=None,
+            username=None,
+            password=None,
+            # Similar to previous options, set it False will use mqtt5 setup for
+            # websockets. It is not necessary means the websocket is disabled.
+            use_websockets=False,
+            websocket_proxy_options=None,
+            websocket_handshake_transform=None,
+            proxy_options=None
+        )
